package sun.nio.ch;

import java.util.Collections;
import sun.net.ExtendedSocketOptions;
import java.util.HashSet;
import java.nio.channels.NetworkChannel;
import java.nio.channels.NoConnectionPendingException;
import java.net.InetAddress;
import java.nio.channels.AlreadyConnectedException;
import java.nio.channels.ConnectionPendingException;
import java.nio.channels.AlreadyBoundException;
import sun.net.NetHooks;
import java.nio.channels.AsynchronousCloseException;
import java.nio.ByteBuffer;
import java.nio.channels.NotYetConnectedException;
import java.util.Set;
import java.net.StandardProtocolFamily;
import java.net.StandardSocketOptions;
import java.net.SocketOption;
import java.nio.channels.ClosedChannelException;
import java.net.SocketAddress;
import java.io.IOException;
import java.nio.channels.spi.SelectorProvider;
import java.net.Socket;
import java.net.InetSocketAddress;
import java.io.FileDescriptor;
import java.nio.channels.SocketChannel;

class SocketChannelImpl extends SocketChannel implements SelChImpl
{
    private static NativeDispatcher nd;
    private final FileDescriptor fd;
    private final int fdVal;
    private volatile long readerThread;
    private volatile long writerThread;
    private final Object readLock;
    private final Object writeLock;
    private final Object stateLock;
    private boolean isReuseAddress;
    private static final int ST_UNINITIALIZED = -1;
    private static final int ST_UNCONNECTED = 0;
    private static final int ST_PENDING = 1;
    private static final int ST_CONNECTED = 2;
    private static final int ST_KILLPENDING = 3;
    private static final int ST_KILLED = 4;
    private int state;
    private InetSocketAddress localAddress;
    private InetSocketAddress remoteAddress;
    private boolean isInputOpen;
    private boolean isOutputOpen;
    private boolean readyToConnect;
    private Socket socket;
    
    SocketChannelImpl(SelectorProvider provider) throws IOException {
        super(provider);
        this.readerThread = 0L;
        this.writerThread = 0L;
        this.readLock = new Object();
        this.writeLock = new Object();
        this.stateLock = new Object();
        this.state = -1;
        this.isInputOpen = true;
        this.isOutputOpen = true;
        this.readyToConnect = false;
        this.fd = Net.socket(true);
        this.fdVal = IOUtil.fdVal(this.fd);
        this.state = 0;
    }
    
    SocketChannelImpl(SelectorProvider provider, FileDescriptor fd, boolean b) throws IOException {
        super(provider);
        this.readerThread = 0L;
        this.writerThread = 0L;
        this.readLock = new Object();
        this.writeLock = new Object();
        this.stateLock = new Object();
        this.state = -1;
        this.isInputOpen = true;
        this.isOutputOpen = true;
        this.readyToConnect = false;
        this.fd = fd;
        this.fdVal = IOUtil.fdVal(fd);
        this.state = 0;
        if (b) {
            this.localAddress = Net.localAddress(fd);
        }
    }
    
    SocketChannelImpl(SelectorProvider provider, FileDescriptor fd, InetSocketAddress remoteAddress) throws IOException {
        super(provider);
        this.readerThread = 0L;
        this.writerThread = 0L;
        this.readLock = new Object();
        this.writeLock = new Object();
        this.stateLock = new Object();
        this.state = -1;
        this.isInputOpen = true;
        this.isOutputOpen = true;
        this.readyToConnect = false;
        this.fd = fd;
        this.fdVal = IOUtil.fdVal(fd);
        this.state = 2;
        this.localAddress = Net.localAddress(fd);
        this.remoteAddress = remoteAddress;
    }
    
    @Override
    public Socket socket() {
        synchronized (this.stateLock) {
            if (this.socket == null) {
                this.socket = SocketAdaptor.create(this);
            }
            return this.socket;
        }
    }
    
    @Override
    public SocketAddress getLocalAddress() throws IOException {
        synchronized (this.stateLock) {
            if (!this.isOpen()) {
                throw new ClosedChannelException();
            }
            return Net.getRevealedLocalAddress(this.localAddress);
        }
    }
    
    @Override
    public SocketAddress getRemoteAddress() throws IOException {
        synchronized (this.stateLock) {
            if (!this.isOpen()) {
                throw new ClosedChannelException();
            }
            return this.remoteAddress;
        }
    }
    
    @Override
    public <T> SocketChannel setOption(SocketOption<T> obj, T t) throws IOException {
        if (obj == null) {
            throw new NullPointerException();
        }
        if (!this.supportedOptions().contains(obj)) {
            throw new UnsupportedOperationException("'" + obj + "' not supported");
        }
        synchronized (this.stateLock) {
            if (!this.isOpen()) {
                throw new ClosedChannelException();
            }
            if (obj == StandardSocketOptions.IP_TOS) {
                Net.setSocketOption(this.fd, Net.isIPv6Available() ? StandardProtocolFamily.INET6 : StandardProtocolFamily.INET, obj, t);
                return this;
            }
            if (obj == StandardSocketOptions.SO_REUSEADDR && Net.useExclusiveBind()) {
                this.isReuseAddress = (boolean)t;
                return this;
            }
            Net.setSocketOption(this.fd, Net.UNSPEC, obj, t);
            return this;
        }
    }
    
    @Override
    public <T> T getOption(SocketOption<T> obj) throws IOException {
        if (obj == null) {
            throw new NullPointerException();
        }
        if (!this.supportedOptions().contains(obj)) {
            throw new UnsupportedOperationException("'" + obj + "' not supported");
        }
        synchronized (this.stateLock) {
            if (!this.isOpen()) {
                throw new ClosedChannelException();
            }
            if (obj == StandardSocketOptions.SO_REUSEADDR && Net.useExclusiveBind()) {
                return Boolean.valueOf(this.isReuseAddress);
            }
            if (obj == StandardSocketOptions.IP_TOS) {
                return Net.getSocketOption(this.fd, Net.isIPv6Available() ? StandardProtocolFamily.INET6 : StandardProtocolFamily.INET, obj);
            }
            return Net.getSocketOption(this.fd, Net.UNSPEC, obj);
        }
    }
    
    @Override
    public final Set<SocketOption<?>> supportedOptions() {
        return SocketChannelImpl.DefaultOptionsHolder.defaultOptions;
    }
    
    private boolean ensureReadOpen() throws ClosedChannelException {
        synchronized (this.stateLock) {
            if (!this.isOpen()) {
                throw new ClosedChannelException();
            }
            if (!this.isConnected()) {
                throw new NotYetConnectedException();
            }
            if (!this.isInputOpen) {
                return false;
            }
            return true;
        }
    }
    
    private void ensureWriteOpen() throws ClosedChannelException {
        synchronized (this.stateLock) {
            if (!this.isOpen()) {
                throw new ClosedChannelException();
            }
            if (!this.isOutputOpen) {
                throw new ClosedChannelException();
            }
            if (!this.isConnected()) {
                throw new NotYetConnectedException();
            }
        }
    }
    
    private void readerCleanup() throws IOException {
        synchronized (this.stateLock) {
            this.readerThread = 0L;
            if (this.state == 3) {
                this.kill();
            }
        }
    }
    
    private void writerCleanup() throws IOException {
        synchronized (this.stateLock) {
            this.writerThread = 0L;
            if (this.state == 3) {
                this.kill();
            }
        }
    }
    
    @Override
    public int read(ByteBuffer byteBuffer) throws IOException {
        if (byteBuffer == null) {
            throw new NullPointerException();
        }
        synchronized (this.readLock) {
            if (!this.ensureReadOpen()) {
                return -1;
            }
            int read = 0;
            try {
                this.begin();
                synchronized (this.stateLock) {
                    if (!this.isOpen()) {
                        int n = 0;
                        this.readerCleanup();
                        this.end(read > 0 || read == -2);
                        try {
                            if (read <= 0 && !this.isInputOpen) {
                                int n2 = -1;
                                return n2;
                            }
                        }
                        finally {}
                        assert !(!IOStatus.check(read));
                        return n;
                    }
                    else {
                        this.readerThread = NativeThread.current();
                    }
                }
                while (true) {
                    read = IOUtil.read(this.fd, byteBuffer, -1L, SocketChannelImpl.nd);
                    if (read == -3 && this.isOpen()) {
                        continue;
                    }
                    break;
                }
                int normalize = IOStatus.normalize(read);
                this.readerCleanup();
                this.end(read > 0 || read == -2);
                synchronized (this.stateLock) {
                    if (read <= 0 && !this.isInputOpen) {
                        int n3 = -1;
                        return n3;
                    }
                }
                assert !(!IOStatus.check(read));
                return normalize;
            }
            finally {
                this.readerCleanup();
                this.end(read > 0 || read == -2);
                synchronized (this.stateLock) {
                    if (read <= 0 && !this.isInputOpen) {
                        int n4 = -1;
                        return n4;
                    }
                }
                assert !(!IOStatus.check(read));
            }
        }
    }
    
    @Override
    public long read(ByteBuffer[] array, int n, int n2) throws IOException {
        if (n < 0 || n2 < 0 || n > array.length - n2) {
            throw new IndexOutOfBoundsException();
        }
        synchronized (this.readLock) {
            if (!this.ensureReadOpen()) {
                return -1L;
            }
            long read = 0L;
            try {
                this.begin();
                synchronized (this.stateLock) {
                    if (!this.isOpen()) {
                        long n3 = 0L;
                        this.readerCleanup();
                        this.end(read > 0L || read == -2L);
                        try {
                            if (read <= 0L && !this.isInputOpen) {
                                long n4 = -1L;
                                return n4;
                            }
                        }
                        finally {}
                        assert !(!IOStatus.check(read));
                        return n3;
                    }
                    else {
                        this.readerThread = NativeThread.current();
                    }
                }
                while (true) {
                    read = IOUtil.read(this.fd, array, n, n2, SocketChannelImpl.nd);
                    if (read == -3L && this.isOpen()) {
                        continue;
                    }
                    break;
                }
                long normalize = IOStatus.normalize(read);
                this.readerCleanup();
                this.end(read > 0L || read == -2L);
                synchronized (this.stateLock) {
                    if (read <= 0L && !this.isInputOpen) {
                        long n5 = -1L;
                        return n5;
                    }
                }
                assert !(!IOStatus.check(read));
                return normalize;
            }
            finally {
                this.readerCleanup();
                this.end(read > 0L || read == -2L);
                synchronized (this.stateLock) {
                    if (read <= 0L && !this.isInputOpen) {
                        long n6 = -1L;
                        return n6;
                    }
                }
                assert !(!IOStatus.check(read));
            }
        }
    }
    
    @Override
    public int write(ByteBuffer byteBuffer) throws IOException {
        if (byteBuffer == null) {
            throw new NullPointerException();
        }
        synchronized (this.writeLock) {
            this.ensureWriteOpen();
            int write = 0;
            try {
                this.begin();
                synchronized (this.stateLock) {
                    if (!this.isOpen()) {
                        int n = 0;
                        this.writerCleanup();
                        this.end(write > 0 || write == -2);
                        try {
                            if (write <= 0 && !this.isOutputOpen) {
                                throw new AsynchronousCloseException();
                            }
                        }
                        finally {}
                        assert !(!IOStatus.check(write));
                        return n;
                    }
                    else {
                        this.writerThread = NativeThread.current();
                    }
                }
                while (true) {
                    write = IOUtil.write(this.fd, byteBuffer, -1L, SocketChannelImpl.nd);
                    if (write == -3 && this.isOpen()) {
                        continue;
                    }
                    break;
                }
                int normalize = IOStatus.normalize(write);
                this.writerCleanup();
                this.end(write > 0 || write == -2);
                synchronized (this.stateLock) {
                    if (write <= 0 && !this.isOutputOpen) {
                        throw new AsynchronousCloseException();
                    }
                }
                assert !(!IOStatus.check(write));
                return normalize;
            }
            finally {
                this.writerCleanup();
                this.end(write > 0 || write == -2);
                synchronized (this.stateLock) {
                    if (write <= 0 && !this.isOutputOpen) {
                        throw new AsynchronousCloseException();
                    }
                }
                assert !(!IOStatus.check(write));
            }
        }
    }
    
    @Override
    public long write(ByteBuffer[] array, int n, int n2) throws IOException {
        if (n < 0 || n2 < 0 || n > array.length - n2) {
            throw new IndexOutOfBoundsException();
        }
        synchronized (this.writeLock) {
            this.ensureWriteOpen();
            long write = 0L;
            try {
                this.begin();
                synchronized (this.stateLock) {
                    if (!this.isOpen()) {
                        long n3 = 0L;
                        this.writerCleanup();
                        this.end(write > 0L || write == -2L);
                        try {
                            if (write <= 0L && !this.isOutputOpen) {
                                throw new AsynchronousCloseException();
                            }
                        }
                        finally {}
                        assert !(!IOStatus.check(write));
                        return n3;
                    }
                    else {
                        this.writerThread = NativeThread.current();
                    }
                }
                while (true) {
                    write = IOUtil.write(this.fd, array, n, n2, SocketChannelImpl.nd);
                    if (write == -3L && this.isOpen()) {
                        continue;
                    }
                    break;
                }
                long normalize = IOStatus.normalize(write);
                this.writerCleanup();
                this.end(write > 0L || write == -2L);
                synchronized (this.stateLock) {
                    if (write <= 0L && !this.isOutputOpen) {
                        throw new AsynchronousCloseException();
                    }
                }
                assert !(!IOStatus.check(write));
                return normalize;
            }
            finally {
                this.writerCleanup();
                this.end(write > 0L || write == -2L);
                synchronized (this.stateLock) {
                    if (write <= 0L && !this.isOutputOpen) {
                        throw new AsynchronousCloseException();
                    }
                }
                assert !(!IOStatus.check(write));
            }
        }
    }
    
    int sendOutOfBandData(byte b) throws IOException {
        synchronized (this.writeLock) {
            this.ensureWriteOpen();
            int sendOutOfBandData = 0;
            try {
                this.begin();
                synchronized (this.stateLock) {
                    if (!this.isOpen()) {
                        int n = 0;
                        this.writerCleanup();
                        this.end(sendOutOfBandData > 0 || sendOutOfBandData == -2);
                        try {
                            if (sendOutOfBandData <= 0 && !this.isOutputOpen) {
                                throw new AsynchronousCloseException();
                            }
                        }
                        finally {}
                        assert !(!IOStatus.check(sendOutOfBandData));
                        return n;
                    }
                    else {
                        this.writerThread = NativeThread.current();
                    }
                }
                while (true) {
                    sendOutOfBandData = sendOutOfBandData(this.fd, b);
                    if (sendOutOfBandData == -3 && this.isOpen()) {
                        continue;
                    }
                    break;
                }
                int normalize = IOStatus.normalize(sendOutOfBandData);
                this.writerCleanup();
                this.end(sendOutOfBandData > 0 || sendOutOfBandData == -2);
                synchronized (this.stateLock) {
                    if (sendOutOfBandData <= 0 && !this.isOutputOpen) {
                        throw new AsynchronousCloseException();
                    }
                }
                assert !(!IOStatus.check(sendOutOfBandData));
                return normalize;
            }
            finally {
                this.writerCleanup();
                this.end(sendOutOfBandData > 0 || sendOutOfBandData == -2);
                synchronized (this.stateLock) {
                    if (sendOutOfBandData <= 0 && !this.isOutputOpen) {
                        throw new AsynchronousCloseException();
                    }
                }
                assert !(!IOStatus.check(sendOutOfBandData));
            }
        }
    }
    
    @Override
    protected void implConfigureBlocking(boolean b) throws IOException {
        IOUtil.configureBlocking(this.fd, b);
    }
    
    public InetSocketAddress localAddress() {
        synchronized (this.stateLock) {
            return this.localAddress;
        }
    }
    
    public SocketAddress remoteAddress() {
        synchronized (this.stateLock) {
            return this.remoteAddress;
        }
    }
    
    @Override
    public SocketChannel bind(SocketAddress socketAddress) throws IOException {
        synchronized (this.readLock) {
            synchronized (this.writeLock) {
                synchronized (this.stateLock) {
                    if (!this.isOpen()) {
                        throw new ClosedChannelException();
                    }
                    if (this.state == 1) {
                        throw new ConnectionPendingException();
                    }
                    if (this.localAddress != null) {
                        throw new AlreadyBoundException();
                    }
                    InetSocketAddress inetSocketAddress = (socketAddress == null) ? new InetSocketAddress(0) : Net.checkAddress(socketAddress);
                    SecurityManager securityManager = System.getSecurityManager();
                    if (securityManager != null) {
                        securityManager.checkListen(inetSocketAddress.getPort());
                    }
                    NetHooks.beforeTcpBind(this.fd, inetSocketAddress.getAddress(), inetSocketAddress.getPort());
                    Net.bind(this.fd, inetSocketAddress.getAddress(), inetSocketAddress.getPort());
                    this.localAddress = Net.localAddress(this.fd);
                }
            }
        }
        return this;
    }
    
    @Override
    public boolean isConnected() {
        synchronized (this.stateLock) {
            return this.state == 2;
        }
    }
    
    @Override
    public boolean isConnectionPending() {
        synchronized (this.stateLock) {
            return this.state == 1;
        }
    }
    
    void ensureOpenAndUnconnected() throws IOException {
        synchronized (this.stateLock) {
            if (!this.isOpen()) {
                throw new ClosedChannelException();
            }
            if (this.state == 2) {
                throw new AlreadyConnectedException();
            }
            if (this.state == 1) {
                throw new ConnectionPendingException();
            }
        }
    }
    
    @Override
    public boolean connect(SocketAddress socketAddress) throws IOException {
        synchronized (this.readLock) {
            synchronized (this.writeLock) {
                this.ensureOpenAndUnconnected();
                InetSocketAddress checkAddress = Net.checkAddress(socketAddress);
                SecurityManager securityManager = System.getSecurityManager();
                if (securityManager != null) {
                    securityManager.checkConnect(checkAddress.getAddress().getHostAddress(), checkAddress.getPort());
                }
                synchronized (this.blockingLock()) {
                    int connect = 0;
                    try {
                        try {
                            this.begin();
                            synchronized (this.stateLock) {
                                if (!this.isOpen()) {
                                    boolean n = false;
                                    this.readerCleanup();
                                    this.end(connect > 0 || connect == -2);
                                    assert !(!IOStatus.check(connect));
                                    return n;
                                }
                                else {
                                    if (this.localAddress == null) {
                                        NetHooks.beforeTcpConnect(this.fd, checkAddress.getAddress(), checkAddress.getPort());
                                    }
                                    this.readerThread = NativeThread.current();
                                }
                            }
                            while (true) {
                                InetAddress inetAddress = checkAddress.getAddress();
                                if (inetAddress.isAnyLocalAddress()) {
                                    inetAddress = InetAddress.getLocalHost();
                                }
                                connect = Net.connect(this.fd, inetAddress, checkAddress.getPort());
                                if (connect == -3 && this.isOpen()) {
                                    continue;
                                }
                                break;
                            }
                        }
                        finally {
                            this.readerCleanup();
                            this.end(connect > 0 || connect == -2);
                            assert !(!IOStatus.check(connect));
                        }
                    }
                    catch (IOException ex) {
                        this.close();
                        throw ex;
                    }
                    synchronized (this.stateLock) {
                        this.remoteAddress = checkAddress;
                        if (connect > 0) {
                            this.state = 2;
                            if (this.isOpen()) {
                                this.localAddress = Net.localAddress(this.fd);
                            }
                            boolean b = true;
                            return b;
                        }
                        else if (!this.isBlocking()) {
                            this.state = 1;
                        }
                        else
                            assert false;
                    }
                }
                boolean b2 = false;
                return b2;
            }
        }
    }
    
    @Override
    public boolean finishConnect() throws IOException {
        synchronized (this.readLock) {
            synchronized (this.writeLock) {
                synchronized (this.stateLock) {
                    if (!this.isOpen()) {
                        throw new ClosedChannelException();
                    }
                    if (this.state == 2) {
                        boolean b = true;
                        return b;
                    }
                    if (this.state != 1) {
                        throw new NoConnectionPendingException();
                    }
                }
                int n = 0;
                try {
                    try {
                        this.begin();
                        synchronized (this.blockingLock()) {
                            synchronized (this.stateLock) {
                                if (!this.isOpen()) {
                                    boolean n2 = false;
                                    try {
                                        this.readerThread = 0L;
                                        if (this.state == 3) {
                                            this.kill();
                                            n = 0;
                                        }
                                    }
                                    finally {}
                                    this.end(n > 0 || n == -2);
                                    assert !(!IOStatus.check(n));
                                    return n2;
                                }
                                else {
                                    this.readerThread = NativeThread.current();
                                }
                            }
                            if (!this.isBlocking()) {
                                while (true) {
                                    n = checkConnect(this.fd, false, this.readyToConnect);
                                    if (n == -3 && this.isOpen()) {
                                        continue;
                                    }
                                    break;
                                }
                            }
                            else {
                                while (true) {
                                    n = checkConnect(this.fd, true, this.readyToConnect);
                                    if (n == 0) {
                                        continue;
                                    }
                                    if (n == -3 && this.isOpen()) {
                                        continue;
                                    }
                                    break;
                                }
                            }
                        }
                    }
                    finally {
                        synchronized (this.stateLock) {
                            this.readerThread = 0L;
                            if (this.state == 3) {
                                this.kill();
                                n = 0;
                            }
                        }
                        this.end(n > 0 || n == -2);
                        assert !(!IOStatus.check(n));
                    }
                }
                catch (IOException ex) {
                    this.close();
                    throw ex;
                }
                if (n > 0) {
                    synchronized (this.stateLock) {
                        this.state = 2;
                        if (this.isOpen()) {
                            this.localAddress = Net.localAddress(this.fd);
                        }
                    }
                    boolean b3 = true;
                    return b3;
                }
                boolean b2 = false;
                return b2;
            }
        }
    }
    
    @Override
    public SocketChannel shutdownInput() throws IOException {
        synchronized (this.stateLock) {
            if (!this.isOpen()) {
                throw new ClosedChannelException();
            }
            if (!this.isConnected()) {
                throw new NotYetConnectedException();
            }
            if (this.isInputOpen) {
                Net.shutdown(this.fd, 0);
                if (this.readerThread != 0L) {
                    NativeThread.signal(this.readerThread);
                }
                this.isInputOpen = false;
            }
            return this;
        }
    }
    
    @Override
    public SocketChannel shutdownOutput() throws IOException {
        synchronized (this.stateLock) {
            if (!this.isOpen()) {
                throw new ClosedChannelException();
            }
            if (!this.isConnected()) {
                throw new NotYetConnectedException();
            }
            if (this.isOutputOpen) {
                Net.shutdown(this.fd, 1);
                if (this.writerThread != 0L) {
                    NativeThread.signal(this.writerThread);
                }
                this.isOutputOpen = false;
            }
            return this;
        }
    }
    
    public boolean isInputOpen() {
        synchronized (this.stateLock) {
            return this.isInputOpen;
        }
    }
    
    public boolean isOutputOpen() {
        synchronized (this.stateLock) {
            return this.isOutputOpen;
        }
    }
    
    @Override
    protected void implCloseSelectableChannel() throws IOException {
        synchronized (this.stateLock) {
            this.isInputOpen = false;
            this.isOutputOpen = false;
            if (this.state != 4) {
                SocketChannelImpl.nd.preClose(this.fd);
            }
            if (this.readerThread != 0L) {
                NativeThread.signal(this.readerThread);
            }
            if (this.writerThread != 0L) {
                NativeThread.signal(this.writerThread);
            }
            if (!this.isRegistered()) {
                this.kill();
            }
        }
    }
    
    @Override
    public void kill() throws IOException {
        synchronized (this.stateLock) {
            if (this.state == 4) {
                return;
            }
            if (this.state == -1) {
                this.state = 4;
                return;
            }
            assert !this.isOpen() && !this.isRegistered();
            if (this.readerThread == 0L && this.writerThread == 0L) {
                SocketChannelImpl.nd.close(this.fd);
                this.state = 4;
            }
            else {
                this.state = 3;
            }
        }
    }
    
    public boolean translateReadyOps(int n, int n2, SelectionKeyImpl selectionKeyImpl) {
        int nioInterestOps = selectionKeyImpl.nioInterestOps();
        int nioReadyOps = selectionKeyImpl.nioReadyOps();
        int n3 = n2;
        if ((n & Net.POLLNVAL) != 0x0) {
            return false;
        }
        if ((n & (Net.POLLERR | Net.POLLHUP)) != 0x0) {
            selectionKeyImpl.nioReadyOps(nioInterestOps);
            this.readyToConnect = true;
            return (nioInterestOps & ~nioReadyOps) != 0x0;
        }
        if ((n & Net.POLLIN) != 0x0 && (nioInterestOps & 0x1) != 0x0 && this.state == 2) {
            n3 |= 0x1;
        }
        if ((n & Net.POLLCONN) != 0x0 && (nioInterestOps & 0x8) != 0x0 && (this.state == 0 || this.state == 1)) {
            n3 |= 0x8;
            this.readyToConnect = true;
        }
        if ((n & Net.POLLOUT) != 0x0 && (nioInterestOps & 0x4) != 0x0 && this.state == 2) {
            n3 |= 0x4;
        }
        selectionKeyImpl.nioReadyOps(n3);
        return (n3 & ~nioReadyOps) != 0x0;
    }
    
    @Override
    public boolean translateAndUpdateReadyOps(int n, SelectionKeyImpl selectionKeyImpl) {
        return this.translateReadyOps(n, selectionKeyImpl.nioReadyOps(), selectionKeyImpl);
    }
    
    @Override
    public boolean translateAndSetReadyOps(int n, SelectionKeyImpl selectionKeyImpl) {
        return this.translateReadyOps(n, 0, selectionKeyImpl);
    }
    
    int poll(int n, long n2) throws IOException {
        assert Thread.holdsLock(this.blockingLock()) && !this.isBlocking();
        synchronized (this.readLock) {
            int poll = 0;
            try {
                this.begin();
                synchronized (this.stateLock) {
                    if (!this.isOpen()) {
                        int n3 = 0;
                        this.readerCleanup();
                        this.end(poll > 0);
                        return n3;
                    }
                    this.readerThread = NativeThread.current();
                }
                poll = Net.poll(this.fd, n, n2);
            }
            finally {
                this.readerCleanup();
                this.end(poll > 0);
            }
            return poll;
        }
    }
    
    @Override
    public void translateAndSetInterestOps(int n, SelectionKeyImpl selectionKeyImpl) {
        int n2 = 0;
        if ((n & 0x1) != 0x0) {
            n2 |= Net.POLLIN;
        }
        if ((n & 0x4) != 0x0) {
            n2 |= Net.POLLOUT;
        }
        if ((n & 0x8) != 0x0) {
            n2 |= Net.POLLCONN;
        }
        selectionKeyImpl.selector.putEventOps(selectionKeyImpl, n2);
    }
    
    @Override
    public FileDescriptor getFD() {
        return this.fd;
    }
    
    @Override
    public int getFDVal() {
        return this.fdVal;
    }
    
    @Override
    public String toString() {
        StringBuffer sb = new StringBuffer();
        sb.append(this.getClass().getSuperclass().getName());
        sb.append('[');
        if (!this.isOpen()) {
            sb.append("closed");
        }
        else {
            synchronized (this.stateLock) {
                switch (this.state) {
                    case 0: {
                        sb.append("unconnected");
                        break;
                    }
                    case 1: {
                        sb.append("connection-pending");
                        break;
                    }
                    case 2: {
                        sb.append("connected");
                        if (!this.isInputOpen) {
                            sb.append(" ishut");
                        }
                        if (!this.isOutputOpen) {
                            sb.append(" oshut");
                            break;
                        }
                        break;
                    }
                }
                InetSocketAddress localAddress = this.localAddress();
                if (localAddress != null) {
                    sb.append(" local=");
                    sb.append(Net.getRevealedLocalAddressAsString(localAddress));
                }
                if (this.remoteAddress() != null) {
                    sb.append(" remote=");
                    sb.append(this.remoteAddress().toString());
                }
            }
        }
        sb.append(']');
        return sb.toString();
    }
    
    private static native int checkConnect(FileDescriptor p0, boolean p1, boolean p2) throws IOException;
    
    private static native int sendOutOfBandData(FileDescriptor p0, byte p1) throws IOException;
    
    static {
        IOUtil.load();
        SocketChannelImpl.nd = new SocketDispatcher();
    }
    
    private static class DefaultOptionsHolder
    {
        static final Set<SocketOption<?>> defaultOptions;
        
        private static Set<SocketOption<?>> defaultOptions() {
            HashSet s = new HashSet(8);
            s.add(StandardSocketOptions.SO_SNDBUF);
            s.add(StandardSocketOptions.SO_RCVBUF);
            s.add(StandardSocketOptions.SO_KEEPALIVE);
            s.add(StandardSocketOptions.SO_REUSEADDR);
            s.add(StandardSocketOptions.SO_LINGER);
            s.add(StandardSocketOptions.TCP_NODELAY);
            s.add(StandardSocketOptions.IP_TOS);
            s.add(ExtendedSocketOption.SO_OOBINLINE);
            s.addAll(ExtendedSocketOptions.clientSocketOptions());
            return Collections.unmodifiableSet(s);
        }
        
        static {
            defaultOptions = defaultOptions();
        }
    }
}
